import hashlib
import getpass
import struct
from Crypto.Hash import SHA256
from Crypto.Cipher import AES
from Crypto import Random
import seccure
import array
import os

import pickle

def hash_addr(unhashed_addr, key=''):
	hVal = hashlib.sha256(unhashed_addr)
	hVal.update(key)
	return hVal.hexdigest()

def username_to_uid(username):
	uid_str = hashlib.sha256(username).digest()[:4]
	uid = struct.unpack('>L', uid_str)[0]
	return uid

def get_uid():
	username = getpass.getuser()
	return username_to_uid(username)

class CounterGen(object):
	"""
	Object to count n-byte binary data.
	Useful for CTR mode and auto-generating
	ECC private keys based on one seed.
	
	usage is
	x = CounterGen(bytestring seed)
	x.counter() to progressively count 1 by 1
	x.countTo(val) to reset value to original seed and
		then count to val
	"""
	def __init__(self, iv=None):
		if iv is None:
			iv = os.urandom(16)
		self.iv = iv
		self.current = array.array('B', self.iv)
	def counter(self):
		self.current[0] = (self.current[0] + 1) % 256
		i = 0
		while (self.current[i] == 0):
			i += 1
			if (i == len(self.current)):
				i = 0
			self.current[i] = (self.current[i] + 1) % 256
		return self.current.tostring()
	def countTo(self, blockNum):
		self.current = array.array('B', self.iv)
		result = self.current.tostring()
		for i in range(blockNum):
			result = self.counter()
		return result

class CryptoContainer():
	'Crypto module to run files through'
	
	def __init__(self, root='/'):
		self.rnd = Random.new()
		self.username = ""
		self.key = {}
		self.known_pkeys = {}
		self.aes_keys = {}
		self.write_keys = {}
		self.directory_read_keys = {}
		self.directory_write_keys = {}
		self.traversal_keys = {}
		self.data_len = 65536 * 3 / 2
		self.root = root
		self.loadcred()

	def loadcred(self):
		keyfile = open('client_temp_keyfile', 'rb')
		key_plaintext = keyfile.next()
		keypair = key_plaintext.split()
		self.username = keypair[0]
		self.uid = username_to_uid(self.username)
		self.key["pub"] = str(seccure.passphrase_to_pubkey(keypair[1]))
		self.key["pri"] = keypair[1]
		for line in keyfile:
			pair = line.split()
			if len(pair) < 2:
				break
			self.known_pkeys[pair[0]] = pair[1]
		keyfile.close()
		self.aes_keys[self.root] = "0000000000000000"
		self.write_keys[self.root] = "0"
		self.directory_read_keys[self.root] = "0"
		self.directory_write_keys[self.root] = str(seccure.passphrase_to_pubkey("0"))
		self.traversal_keys[self.root] = "0000000000000000"
		
	
	def generateInodeAddress(self, fileName, parent=None):
		"""
		Generate the real address of an inode block

		Keyword arguments:
		parent -- the inode object of the parent directory
		fileName -- the string of the current fileName
		
		return value: an inode's real address on the server
		"""
		if parent is None:
			parent = lambda:0
			parent.name = self.root
		if parent.name not in self.traversal_keys:
			return False
		key = self.traversal_keys[parent.name]
		return hash_addr(fileName, str(key))
	def generateBlockAddress(self, inode, blockNum):
		"""
		Generate the real address of a data block

		Keyword arguments:
		inode -- the inode object of this file
		blockNum -- the integer index of the block to lookup
		
		return value: a block's real address on the server
		"""
		if inode.parent is None:
			name = self.root
		else:
			name = inode.parent.name
		if name not in self.traversal_keys:
			return False
		key = self.traversal_keys[name]
		return hash_addr(inode.name + str(blockNum), str(key))

	
	def newFile(self, inode):
		"""
		Takes care of the associated crypto plans for a new file
		
		Keyword arguments:
		inode -- the inode object of this file
		
		return value: none, assumed to succeed
		"""
		self.aes_keys[inode.name] = self.rnd.read(16)
		self.write_keys[inode.name] = self.rnd.read(16)
		accessList = {"read": set([self.uid]), "write": set([self.uid])}
		self.writeACL(inode, accessList)
	def newDirectory(self, inode):
		"""
		Takes care of the associated crypto plans for a new directory
		
		Keyword arguments:
		inode -- the inode object of this file
		
		return value: none, assumed to succeed
		"""
		self.aes_keys[inode.name] = self.rnd.read(16)
		self.write_keys[inode.name] = str(self.rnd.read(16))
		self.directory_read_keys[inode.name] = str(self.rnd.read(16))
		self.directory_write_keys[inode.name] = str(self.directory_read_keys[inode.name])
		self.traversal_keys[inode.name] = self.rnd.read(16)
		accessList = {"read_blocks": set([self.uid]), "sign_blocks": set([self.uid]), "read_entries": set([self.uid]), 
									"write_entries": set([self.uid]), "traverse": set([self.uid]), }
		self.writeACL(inode, accessList)
		
	
	def computeFileSig(self, block1Contents, orderedBlocks, fileName):
		"""
		Generate the signature for a complete file or directory in preparation for writing.
		
		Usage: block 1 should be encrypted and signed last. Just before that, compute a
		signature using the signatures of all other blocks in the file, and block1's
		binary "data" contents, but not any other meta info.
		
		Keyword arguments:
		block1Contents -- the binary object containing "real" data from block1
		orderedBlocks -- the list containing all SignedBlock objects, including block1
		fileName -- the string of the current fileName
		
		return value: a string containing the complete file signature
		"""
		if fileName not in self.write_keys:
			return False
		hashObj = hashlib.sha256(block1Contents)
		for b in orderedBlocks[1:]:
			hashObj.update(b.signature)
		keyObj = CounterGen(self.write_keys[fileName])
		return seccure.sign(hashObj.hexdigest(), keyObj.countTo(1))
	def verifyFileSig(self, signature, block1Contents, orderedBlocks):
		"""
		Verify the signature for a complete file or directory before you continue reading.
		
		Usage: block 1 should be decrypted and verified first. Just after that, compute a
		hash using the signatures of all other blocks in the file, and block1's
		binary "data" contents, but not any other meta info, then verify the signature of
		that hash.
		
		Keyword arguments:
		signature -- the string of the signature to be verified
		block1Contents -- the binary object containing "real" data from block1
		orderedBlocks -- the list containing all SignedBlock objects, including block1
		
		return value: whether or not the file's signature checks out
		"""
		hashObj = hashlib.sha256(block1Contents)
		for b in orderedBlocks[1:]:
			hashObj.update(b.signature)
		return seccure.verify(hashObj.hexdigest(), signature, orderedBlocks[0].pubkey)
	
	#input: binary object "contents", str "fileName", int "blockNum", str "aes_key", str "ecc_key"
	def _encryptAndSignBlock(self, contents, fileName, blockNum, aes_key, ecc_key):
		encryptOutput = self._aesencrypt(aes_key, contents)
		internalSig = seccure.sign(encryptOutput, str(blockNum) + str(aes_key))
		dataOut = pickle.dumps([encryptOutput, internalSig])
		return (dataOut, seccure.sign(dataOut, ecc_key), str(seccure.passphrase_to_pubkey(ecc_key)))
	#input: SignedBlock "block", str "fileName", int "blockNum", str "aes_key"
	def _decryptAndVerifyBlock(self, block, fileName, blockNum, aes_key):
		if not seccure.verify(block.contents, block.signature, block.pubkey):
			return False
		cipherArray = pickle.loads(block.contents)
		if not seccure.verify(cipherArray[0], cipherArray[1], str(seccure.passphrase_to_pubkey(str(blockNum) + str(aes_key)))):
			return False
		return self._aesdecrypt(aes_key, cipherArray[0])
	
	def encryptAndSignData(self, contents, fileName, blockNum):
		"""
		Generate the signed encryption of a data block

		Keyword arguments:
		contents -- the binary object containing data from the file
		fileName -- the string of the current fileName
		blockNum -- the integer index of the current block
		
		return value: a string encoding the signed block encryption
		"""
		keyCount = CounterGen(self.write_keys[fileName])
		return self._encryptAndSignBlock(contents, fileName, blockNum, self.aes_keys[fileName], keyCount.countTo(blockNum))
	def decryptAndVerifyData(self, block, fileName, blockNum):
		"""
		Retrieve the contents of a data block after verifying it

		Keyword arguments:
		block -- a SignedBlock object containing desired contents
		fileName -- the string of the current fileName
		blockNum -- the integer index of the current block
		
		return value: a binary object containing the block's stored data
		"""
		return self._decryptAndVerifyBlock(block, fileName, blockNum, self.aes_keys[fileName])
	
	def encryptAndSignInode(self, inode, dirName):
		"""
		Generate the signed encryption of an inode block

		Keyword arguments:
		inode -- the dictionary object containing key information for the current file
		dirName -- the name of the parent directory
		
		return value: a string encoding the signed block encryption
		"""
		return self._encryptAndSignBlock(inode, inode.name, 0, self.traversal_keys[dirName], self.key['pri'])
	def decryptAndVerifyInode(self, block, dirName):
		"""
		Retrieve the contents of an inode block after verifying it

		Keyword arguments:
		block -- a SignedBlock object for an inode
		dirName -- the string of the parent directory's name
		
		return value: a dictionary object containing key information for the current file
		"""
		return self._decryptAndVerifyBlock(block, dirName, 0, self.traversal_keys[dirName])
	
	
	def writeACL(self, inode, accessDict):
		"""
		Write the full ACL field for a file inode. This is pre-block-encryption.
		
		Keyword arguments:
		inode -- the inode object for the file
		accessDict -- the dictionary storing all unencrypted file permissions
		
		return value: whether the write was successful
		"""
		if inode.uid and inode.uid != self.uid:
			return False
		if inode.name not in self.aes_keys:
			print "You lost read bits on your own file...."
			return False
		for perms in accessDict:
			if self.uid not in accessDict[perms]:
				accessDict[perms].append(self.uid)
		inode.acl = self._aesencrypt(self.aes_keys[inode.name], accessDict, False)
		return True
	def readACL(self, inode):
		"""
		Read the full ACL field for a file inode. This is post-block-decryption.
		
		Keyword arguments:
		inode -- the inode object for the file
		
		return value: the unencrypted access dictionary for the file
		"""
		if inode.name not in self.aes_keys:
			return False
		return self._aesdecrypt(self.aes_keys[inode.name], inode.acl, False)
	
	
	def addDirEntry(self, contents, dirName, fileName):
		"""
		Add a new entry to a Directory block

		Keyword arguments:
		contents -- the list that contains the current ecc-encrypted contents of the directory block
		dirName -- the string of this directory's name
		fileName -- the string of the fileName to be added
		
		return value: whether the entry was successful
		"""
		if dirName not in self.directory_write_keys:
			return False
		block.content.append(seccure.encrypt(fileName, self.directory_write_keys[dirName]))
		return True
	def removeDirEntryByName(self, contents, dirName, fileName):
		"""
		Remove an entry from a Directory block by name

		Keyword arguments:
		contents -- the list that contains the current ecc-encrypted contents of the directory block
		dirName -- the string of this directory's name
		fileName -- the string of the fileName to be removed
		
		return value: whether the removal was successful
		"""
		blockContent = self.readDirBlock(block, dirName)
		if not blockContent:
			return False
		try:
			index = blockContent.index(fileName)
		except ValueError:
			return False
		return self.removeDirEntryByIndex(block, index)
	def removeDirEntryByIndex(self, contents, index):
		"""
		Remove an entry from a Directory block by index

		Keyword arguments:
		contents -- the list that contains the current ecc-encrypted contents of the directory block
		index -- the index of the file to be removed
		
		return value: whether the removal was successful
		"""
		try:
			block.content.pop(index)
		except IndexError:
			return False
		return True
	def readDirBlock(self, contents, dirName):
		"""
		Read and output the full contents of a directory block

		Keyword arguments:
		contents -- the list that contains the current ecc-encrypted contents of the directory block
		dirName -- the string of this directory's name
		
		return value: a list containing the unencrypted filenames in this directory block
		"""
		if dirName not in self.directory_read_keys:
			return False
		result = []
		for cipher in block.content:
			result.append(seccure.decrypt(cipher, self.directory_read_keys[dirName]))
		return result
		
	
	def readDirInode(self, inode):
		"""
		Read a directory inode and store your new keys

		Keyword arguments:
		inode -- the dictionary object containing key information for the current file
		
		return value: a list of which permissions you have on this directory
		"""
		perms = []
		dirName = inode.name
		readKey = self._eccDecryptKey(inode.read_list)
		if (readKey):
			self.aes_keys[dirName] = readKey
			perms.append("access blocks")
		writeKey = self._eccDecryptKey(inode.write_list)
		if (writeKey):
			self.write_keys[dirName] = writeKey
			perms.append("sign blocks")
		dReadKey = self._eccDecryptKey(inode.dir_read_list)
		if (dReadKey):
			self.directory_read_keys[dirName] = dReadKey
			perms.append("read entries")
		dWriteKey = self._eccDecryptKey(inode.dir_write_list)
		if (dWriteKey):
			self.directory_write_keys[dirName] = dWriteKey
			perms.append("write entries")
		traversalKey = self._eccDecryptKey(inode.execute_list)
		if (traversalKey):
			self.traversal_keys[dirName] = traversalKey
			perms.append("access internal files")
		return perms
	def writeDirInode(self, inode, accessDict):
		"""
		Write the full access list of a directory inode

		Keyword arguments:
		inode -- the dictionary object to contain encrypted key information for the current file
		accessDict -- the dictionary of unencrypted permissions you want in the inode, pulled from
									the AES encryption within an inode
		
		return value: whether or not all keys were successfully encrypted
									(if not, check client for missing public keys)
		"""
		perfect = True
		readList = accessDict["read_blocks"]
		(inode.read_list, temp) = self._eccEncryptKey(self.aes_keys[inode.name], readList)
		if not temp:
			perfect = False
		readList = accessDict["sign_blocks"]
		(inode.write_list, temp) = self._eccEncryptKey(self.write_keys[inode.name], readList)
		if not temp:
			perfect = False
		readList = accessDict["read_entries"]
		(inode.dir_read_list, temp) = self._eccEncryptKey(self.directory_read_keys[inode.name], readList)
		if not temp:
			perfect = False
		readList = accessDict["write_entries"]
		(inode.dir_write_list, temp) = self._eccEncryptKey(self.directory_write_keys[inode.name], readList)
		if not temp:
			perfect = False
		readList = accessDict["traverse"]
		(inode.execute_list, temp) = self._eccEncryptKey(self.traversal_keys[inode.name], readList)
		if not temp:
			perfect = False
		return perfect
	
	def readFileInode(self, inode):
		"""
		Read a file inode and store your new keys

		Keyword arguments:
		inode -- the dictionary object containing key information for the current file
		
		return value: a list of which permissions you have on this file
		"""
		perms = []
		fileName = inode.name
		readKey = self._eccDecryptKey(inode.read_list)
		if (readKey):
			self.aes_keys[fileName] = readKey
			perms.append("read")
		writeKey = self._eccDecryptKey(inode.write_list)
		if (writeKey):
			self.write_keys[fileName] = writeKey
			perms.append("write")
		return perms
	def writeFileInode(self, inode, accessDict):
		"""
		Write the full access list of a file inode

		Keyword arguments:
		inode -- the dictionary object to contain encrypted key information for the current file
		accessDict -- the dictionary of unencrypted permissions you want in the inode
		
		return value: whether or not all keys were successfully encrypted
									(if not, check client for missing public keys)
		"""
		perfect = True
		readList = accessDict["read"]
		(inode.read_list, temp) = self._eccEncryptKey(self.aes_keys[inode.name], readList)
		if not temp:
			perfect = False
		readList = accessDict["write"]
		(inode.write_list, temp) = self._eccEncryptKey(self.write_keys[inode.name], readList)
		if not temp:
			perfect = False
		return perfect
	
	
	def addPubkey(self, user, key):
		"""
		Add a public key to the local user's keyfile

		Keyword arguments:
		user -- the string containing the desired username or uid
		key -- the string containing the desired public key
		
		return value: None, assumed to succeed. Will silently do nothing if key is known.
		"""
		self.known_pkeys[user] = key
		keyfile = open('client_temp_keyfile', 'rb')
		newline = user + " " + key + "\n"
		for line in keyfile:
			if line == newline:
				return
		keyfile.close()
		keyfile = open('client_temp_keyfile', 'ab')
		keyfile.write(newline)
		keyfile.close()
	
	
	#input: str "key", list "userlist"
	def _eccEncryptKey(self, key, userlist):
		ciphertexts = []
		data = pickle.dumps(key)
		perfect = True
		for user in userlist:
			if user not in self.known_pkeys and user != self.uid:
				print "Ask", user, "for their public key! Not encrypting to", user, "."
				perfect = False
			elif user != self.uid:
				ciphertexts.append(seccure.encrypt(data, self.known_pkeys[user]))
			else:
				ciphertexts.append(seccure.encrypt(data, self.key["pub"]))
		return (ciphertexts, perfect)
	#input: list "ciphertexts"
	def _eccDecryptKey(self, ciphertexts):
		for c in ciphertexts:
			try:
				ptext = seccure.decrypt(c, self.key["pri"])
				return pickle.loads(ptext)
			except IntegrityError:
				pass
		return False
	
	
	#input: str "data"
	def _paddata(self, data):
		padLen = self.data_len - len(data) - 1
		data = data + "1" + "0"*padLen
		return data
	#input: str "data"
	def _unpaddata(self, data):
		endIndex = data.rfind("10")
		return data[:endIndex]
	
	#input: str "key", binary object "data"
	def _aesencrypt(self, key, data, pad=True):
		data = pickle.dumps(data)
		if pad:
			data = self._paddata(data)
		iv = self.rnd.read(16)
		ivObj = CounterGen(iv)
		cipher = AES.new(key, AES.MODE_CTR, counter = ivObj.counter)
		return pickle.dumps([iv, cipher.encrypt(data)])
	#input: str "key", str "pickled"
	def _aesdecrypt(self, key, pickled, padded=True):
		cipherTextObj = pickle.loads(pickled)
		iv = cipherTextObj[0]
		ivObj = CounterGen(iv)
		cipher = AES.new(key, AES.MODE_CTR, counter = ivObj.counter)
		data = cipher.decrypt(cipherTextObj[1])
		if padded:
			data = self._unpaddata(data)
		return pickle.loads(data)
