import pickle
from time import time
from crypto import *

BLOCK_SIZE = 32 * 1024

# Thrown when malicious or incorrect behavior is detected (e.g. wrong signatures)
class VerificationException(Exception):
	pass

class SignedBlock(object):

	def __init__(self, addr, contents, pubkey=None, signature=None):
		self.addr = addr
		self.contents = contents
		self.pubkey = pubkey
		self.signature = signature

	def checkSignature(self):
		#TODO: Implement
		return True

MAX_SIG_LEN = 80
# file_length, content_signature, mtime, ctime
CONTENT_HEADER_FMT = "<l%ds2d" % MAX_SIG_LEN
CONTENT_OFFSET = struct.calcsize(CONTENT_HEADER_FMT)

class File(object):
	def __init__(self, name, uid, gid, mode):
		self.parent = None
		self.store = None
		self._blocks = None
		self._file_length = None
		self.content_signature = None
		self.name = name
		self.uid = uid
		self.gid = gid
		self.mode = mode
		self._ctime = self._mtime = time()

	def init(self, store, crypto, parent):
		"""Initializes a new file"""
		self.store = store
		self.crypto = crypto
		self.parent = parent

	@staticmethod
	def read_from(storage, crypto, parent, name):
		# read the block and return a NormalFile or a Directory
		if parent is None:
			unhashed_addr = name
		else:
			unhashed_addr = parent.entry_to_addr(name)
		addr = hash_addr(unhashed_addr)
		block = storage.read(addr)
		if block is None: 
			return None
		else:
			inode = pickle.loads(block.contents)
			inode.parent = parent
			inode.name = name
			inode.store = storage
			inode.crypto = crypto
			return inode

	def write(self):
		self.store.write(SignedBlock(self.addr, pickle.dumps(self)))
		self.write_block(0, self.read_block(0)) # store fields in first block

	def __getstate__(self):
		# Keep only parts that should be pickled
		inode = {}
		for key in ['uid', 'gid', 'mode', 'acl']:
			inode[key] = getattr(self, key)
		return inode

	def __setstate__(self, contents):
		self.__dict__ = contents
		self.parent = self._file_length = self.store = self.name = self._blocks = None
		self.content_signature = self._mtime = self._ctime = None

	def move(self, new_parent, new_name):
		old_parent = self.parent
		old_parent.remove_entry(self.name)
		if old_parent.addr == new_parent.addr:
			new_parent = old_parent #make sure new_parent is up to date
		new_parent.add_entry(new_name)

		self.store.delete(self.addr)
		self.parent = new_parent
		self.name = new_name
		self.ctime = time()

		self.write()
		new_parent.write()
		old_parent.write()

	@property
	def unhashed_addr(self):
		if self.parent is None:
			return self.name
		else:
			return self.parent.entry_to_addr(self.name)

	@property
	def addr(self):
		return hash_addr(self.unhashed_addr)

	def __len__(self):
		if self._file_length is None:
			self._read_contents()
		return self._file_length

	def set_length(self, length):
		if length < len(self):
			# truncate
			old_num_blocks = self.num_blocks
			self._file_length = length
			new_num_blocks = self.num_blocks
			for i in range(new_num_blocks, old_num_blocks):
				# TODO parallel
				self.store.delete(self.block_addr(i))
			if length % BLOCK_SIZE != 0:
				assert length / BLOCK_SIZE == new_num_blocks - 1
				last_block = self.read_block(length / BLOCK_SIZE)[length % BLOCK_SIZE:]
				self.write_block(length / BLOCK_SIZE, last_block)
		elif length > len(self):
			# zero extend
			self[len(self):length] = '\0' * (length - len(self))
		else:
			return # no change
		self._set_length(length)

	def _set_length(self, length):
		self._file_length = length
		self._ctime = self._mtime = time()
		self.write_block(0, self.read_block(0)) # store new length & times

	def __getitem__(self, ix):
		if isinstance(ix, int):
			offset, end = ix, ix + 1
		elif isinstance(ix, slice):
			if ix.step is not None and ix.step != -1:
				raise ix.step
			offset, end = ix.start or 0, ix.stop or len(self)
		else:
			raise TypeError("index must be int or slice")
		if offset == end:
			return ''
		first_block = offset / BLOCK_SIZE
		last_block = (end + BLOCK_SIZE - 1) / BLOCK_SIZE - 1
		blocks = {}
		for i in range(first_block, last_block + 1):
			# really these reads should happen in parallel
			blocks[i] = self.read_block(i)
		if last_block == first_block:
			content = blocks[first_block][offset % BLOCK_SIZE : end % BLOCK_SIZE]
		else:
			content = blocks[first_block][offset % BLOCK_SIZE:] 
			content += ''.join(blocks[i] for i in range(first_block + 1, last_block))
			content += blocks[last_block][:end % BLOCK_SIZE]
		return content

	def __setitem__(self, ix, data):
		if isinstance(ix, int):
			offset, end = ix, ix + 1
		elif isinstance(ix, slice):
			if ix.step is not None and ix.step != -1:
				raise ix.step
			offset = 0 if ix.start is None else ix.start
			end = len(self) if ix.stop is None else ix.stop
		else:
			raise TypeError("index must be int or slice")
		if end - offset != len(data):
			import pdb; pdb.set_trace()
			raise TypeError("bad data len '%s' (%d,%d)" % (str(data), offset, end))
		if offset == end:
			return
		first_block = offset / BLOCK_SIZE
		last_block = (end + BLOCK_SIZE - 1) / BLOCK_SIZE - 1
		blocks = {}
		blocks[first_block] = self.read_block(first_block)
		off = offset % BLOCK_SIZE
		blocks[first_block] = blocks[first_block][:off] + \
				data[:BLOCK_SIZE - off]
		if last_block != first_block:
			blocks[last_block] = self.read_block(last_block)
			blocks[last_block] = data[-(end % BLOCK_SIZE):] + \
					blocks[last_block][BLOCK_SIZE - (end % BLOCK_SIZE):]
			for i in range(first_block + 1, last_block):
				b_off = i - first_block - 1
				d_off = BLOCK_SIZE - off + b_off * BLOCK_SIZE
				blocks[i] = data[d_off: d_off + BLOCK_SIZE]
		self._set_length(max(len(self), end))
		for i in range(first_block, last_block + 1):
			# TODO parallel write
			self.write_block(i, blocks[i])
		self.ctime = self._mtime = time()


	@property
	def num_blocks(self):
		return (len(self) + BLOCK_SIZE - 1) / BLOCK_SIZE

	@property
	def blocks(self):
		if self._blocks is None:
			self._read_contents()
		return self._blocks

	@property
	def mtime(self):
		if self._blocks is None:
			self._read_contents()
		return self._mtime

	@mtime.setter
	def mtime(self, val):
		if self._blocks is None:
			self._read_contents() # make sure a read won't overwrite the new mtime
		self._mtime = val

	@property
	def ctime(self):
		if self._blocks is None:
			self._read_contents()
		return self._ctime

	@mtime.setter
	def ctime(self, val):
		if self._blocks is None:
			self._read_contents() # make sure a read won't overwrite the new ctime
		self._ctime = val


	def _read_contents(self):
		block0 = self.store.read(self.block_addr(0))
		if block0 is None:
			self._file_length = 0
			self.content_signature = None
			self._mtime = self._ctime = time()
		else:
			self._file_length, self.content_signature, self._mtime, self._ctime = \
					struct.unpack(CONTENT_HEADER_FMT, block0.contents[:CONTENT_OFFSET])
		self._blocks = [self.store.read(self.block_addr(i)) for i in range(self.num_blocks)]
		# TODO verify content with signature

	# Deletes the inode and (TODO) the blocks it points to
	def delete(self):
		self.parent.remove_entry(self.name)
		self.parent.write()
		self.store.delete(self.addr)

	def block_addr(self, index):
		return hash_addr(str(index) + '\0' + self.unhashed_addr)

	# Returns a string of length at most BLOCK_SIZE
	def read_block(self, index):
		if index >= len(self.blocks):
			return ''
		else:
			contents = self.blocks[index].contents
			if index == 0:
				return contents[CONTENT_OFFSET:]
			else:
				return contents

	# block_contents = a string of length at most BLOCK_SIZE
	def write_block(self, index, block_contents):
		if index > len(self.blocks):
			raise "out of bounds"
		if index == 0:
			sig = self.content_signature or ''
			header = struct.pack(
					CONTENT_HEADER_FMT,
					self._file_length, sig, self.mtime, self.ctime
			)
			block_contents = header + block_contents
		block = SignedBlock(self.block_addr(index), block_contents)
		if index == len(self.blocks):
			self.blocks.append(block)
		else:
			self.blocks[index] = block
		self.store.write(block)

class NormalFile(File):
	def init(self, *args):
		File.init(self, *args)
		self.crypto.newFile(self)

class Directory(File):

	def __init__(self, *args):
		File.__init__(self, *args)
		self._entries = None

	def init(self, *args):
		File.init(self, *args)
		self.crypto.newDirectory(self)

	def __setstate__(self, contents):
		File.__setstate__(self, contents)
		self._entries = None

	@property
	def entries(self):
		if self._entries is None:
			self._entries = [e for e in self[:].split('\0') if e]
		return self._entries

	# entry = the name of the new entry
	def add_entry(self, entry):
		self.ctime = self.mtime = time()
		self.entries.append(entry)
		self.write_entries()

	# entry = the name of the entry to remove
	def remove_entry(self, entry):
		self.ctime = self.mtime = time()
		self.entries.remove(entry)
		self.write_entries()

	def write_entries(self):
		self.num_entries = len(self.entries)
		content = '\0'.join(self.entries)
		self[:len(content)] = content

	# entry = the name of a potential entry in the directory
	# Returns the *unhashed* address of the entry's inode -- hash it to get the
	#   address.
	def entry_to_addr(self, entry):
		return self.addr + '/' + entry

	def read_child(self, entry):
		return File.read_from(self.store, self.crypto, self, entry)
